"""
This Python package contains engines that contains the minimum API to connect
an external database engine to a Populse database.
"""


class Engine:
    """
    Base class for database engines. All methods of this base class raise
    NotImplementedError. This class exists to list all the methods that
    must be implemented by an engine.
    """

    def __init__(self):
        """
        Parameters passed to __init__ may differ between engine classes.
        It is up to engine_factory to extract appropriate parameter(s)
        from URL.
        """
        raise NotImplementedError()

    def __enter__(self):
        """
        This method is called at the beginning of a database modification
        session before any other method (normally within a "with" statement).
        Typically, it creates a connection with the database system, starts a
        session then checks the existence of the base schema and creates it if
        necessary.
        """
        raise NotImplementedError()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        This method is called when the user do not need to use the engine
        anymore. After this call, all database resources are freed and a
        call to __enter__ is necessary to be able to reuse the engine.
        Parameters are the exception information as for a "with" statement.
        It must call commit() if exc_type is None or else rollback().
        """
        raise NotImplementedError()

    def commit(self):
        """
        Store in the database all modifications done since the last call to
        either __enter__, commit or rollback.
        """
        raise NotImplementedError()

    def rollback(self):
        """
        Store discards all database modifications done since the last call to
        either __enter__, commit or rollback.
        """
        raise NotImplementedError()

    def clear(self):
        """
        Erase database data and schema.
        """
        raise NotImplementedError()

    def has_collection(self, collection):
        """
        Checks existence of a collection. May be called often,
        must be fast.

        :param collection: collection name (str)
        """
        raise NotImplementedError()

    def add_collection(self, collection, primary_key):
        """
        Create a new collection given the name of a field that will be created
        with a text type. This field is the primary key of each document. It
        means that, in a single collection, a value must uniquely identify a
        document. In other word, two documents cannot have the same value for
        this field.

        :param collection: collection name (str)

        :param primary_key: name of the primary key field (str)
        """
        raise NotImplementedError()

    def collection(self, collection):
        """
        Returns a Row Object with at least the following items:
        collection_name: name of the collection
        primary_key: name of the primary key field
        More engine specific items may be present but they must
        not be used in the general API.
        Returns None if the collection does not exist.

        :param collection: collection name (str)
        """
        raise NotImplementedError()

    def primary_key(self, collection):
        """
        Return the name of the primary key of a collection.
        Returns None if the collection does not exist.

        :param collection: collection name (str)
        """
        raise NotImplementedError()

    def remove_collection(self, collection):
        """
        Delete a collection and its data.

        :param collection: collection name (str)
        """
        raise NotImplementedError()

    def collections(self):
        """
        Return an iterable (e.g. list, generator, etc.) browsing the
        collections presents in the database. See collection() method
        for the format of each collection object.
        """
        raise NotImplementedError()

    def add_field(self, collection, field, type, description, index):
        """
        Adds a new field in a collection.

        :param collection: collection name (str, must be existing)
        :param field: new field name (str, must not be existing)
        :param type: field type, in ('string', 'int', 'float',
           'boolean', 'date', 'datetime', 'time', 'json', 'list_string',
           'list_int', 'list_float', 'list_boolean', 'list_date',
           'list_datetime', 'list_time', 'list_json')
        :param description: field description (str or None)
        :param index: boolean indicating if a database index must be created
           for this field to speed-ud queries involving values for this field.
        """
        raise NotImplementedError()

    def has_field(self, collection, field):
        """
        Checks existence of a field in a collection. May be called often,
        must be fast.
        Returns False if collection does not exists.

        :param collection: collection name (str)

        :param field: field name (str)
        """
        raise NotImplementedError()

    def field(self, collection, field):
        """
        Returns a Row Object corresponding to a collection field with at
        least the following items:
        collection_name: name of the collection
        field_name: field name
        field_type: type of values for this field
        description: text describing the field usage
        has_index: boolean indicating if an index was created for this field
        More engine specific items may be present but they must
        not be used in the general API.
        Returns None if the collection or the field does not exist.

        :param collection: collection name (str)

        :param field: field name (str)
        """
        raise NotImplementedError()

    def fields(self, collection=None):
        """
        Return an iterable (e.g. list, generator, etc.) browsing the
        fields presents in the given collection (or all collections if
        collection parameter is None). See field() method
        for the format of each field object.

        :param collection: collection name (str)
        """
        raise NotImplementedError()

    def remove_fields(self, collection, fields):
        """
        Remove given fields from a collection as well as all corresponding data.

        :param collection: collection name (str)

        :param fields: field name (str)
        """
        raise NotImplementedError()

    def has_document(self, collection, document_id):
        """
        Checks existence of a document in a collection.

        :param collection: collection name (str)

        :param document_id: document identifier: the value of the primary
             key field (str)

        """
        raise NotImplementedError()

    def document(self, collection, document_id, fields=None, as_list=False):
        """
        Returns a Row Object corresponding to a document in the collection.
        The object has one item per selected fields. If fields is not given,
        fields returned by the fields() method are used.
        If as_list is True, a list of values is returned (one value per selected
        field).
        Returns None if the collection or document does not exists.

        :param collection: collection name (str)

        :param document_id: document identifier: the value of the primary
            key field (str)

        :param fields: list of fields to get values from (other fields are
            ignored) (list of str or None)

        :param as_list: if True, return a list of values instead of a Row Object (str)
        """
        raise NotImplementedError()

    def has_value(self, collection, document_id, field):
        """
        Check if a document has a not null value for a given field.
        """
        raise NotImplementedError()

    def set_values(self, collection, document_id, values):
        """
        Change some values in an existing document.

        :param collection: collection name (str)

        :param document_id: document identifier: the value of the primary
            key field (str)

        :param values: dictionary with field/value pairs (dict)
        """
        raise NotImplementedError()

    def remove_value(self, collection, document_id, field):
        """
        Remove a value from a document (setting its value to None).

        :param collection: collection name (str)

        :param document_id: document identifier: the value of the primary
            key field (str)

        :param fields: field name (str)
        """
        raise NotImplementedError()

    def remove_document(self, collection, document_id):
        """
        Remove a document from a collection.

        :param collection: collection name (str)

        :param document_id: document identifier: the value of the primary
            key field (str)
        """
        raise NotImplementedError()

    def parse_filter(self, collection, filter):
        """
        Given a filter string, return a internal query representation that
        can be used with filter_documents() to select documents


        :param collection: the collection for which the filter is intended
               (str, must be existing)

        :param filter: the selection string using the populse_db selection
                       language.
        """
        raise NotImplementedError()

    def filter_documents(
        self, parsed_filter, fields=None, as_list=False, distinct=False
    ):
        """
        Iterate over document selected by a filter. See document() method
        for the format of a document object.

        :param parsed_filter: internal object representing a filter on a
            collection (filter object returned by parse_filter())

        :param fields: list of fields to get values from (other fields are
            ignored) (list of str or None)

        :param as_list: if True, return a list of values instead of a Row Object (str)

        :param distinct: if True, return only a series of different values.
        """
        raise NotImplementedError()



from datetime import datetime

schemas = [
    {
        "version": "1.0.0",
        "schema": {
            # A global value
            "last_update": datetime,
            # A single document (i.e. not in a collection)
            "dataset": {
                "directory": str,
                "schema": "str",
            },
            # A collection of metadata associated to a path.
            "metadata": [
                {
                    "path": [str, {"primary_key": True}],
                    "subject": str,
                    "time_point": str,
                    "history": list[str],  # contains a list of execution_id
                }
            ],
            # A collection of executions to track data provenance
            "execution": [
                {
                    "execution_id": ["str", {"primary_key": True}],
                    "start_time": "datetime",
                    "end_time": datetime,
                    "status": str,
                    "capsul_executable": str,
                    "capsul_parameters": dict,
                    "software": str,
                    "software_module": str,
                    "software_version": str,
                }
            ],
            # A collection of snapshots
            "snapshots": [
                {
                    "subject": ["str", {"primary_key": True}],
                    "time_point": ["str", {"primary_key": True}],
                    "image": str,
                    "top": "list[int]",
                    "size": list[float],
                    "execution": "str",
                    "data_type": "str",
                    "side": "str",
                }
            ],
        },
    }
]



import ast
import datetime

import dateutil.parser
from lark import Lark, Transformer

# The grammar (in Lark format) used to parse filter strings:
filter_grammar = """
?start : filter

?filter : conditions
        | negation
        | "(" filter ")"
        | "(" filter ")" BOOLEAN_OPERATOR filter -> conditions

?conditions : condition (BOOLEAN_OPERATOR filter)*


negation : "NOT"i condition
         | "NOT"i "(" filter ")"

BOOLEAN_OPERATOR : "AND"i
                 | "OR"i

CONDITION_OPERATOR : "=="i
                   | "!="i
                   | "<="i
                   | ">="i
                   | ">"i
                   | "<"i
                   | "IN"i
                   | "ILIKE"i
                   | "LIKE"i

condition : "ALL"i                         -> all
          | operand CONDITION_OPERATOR operand

?operand : literal
         | field_name

field_name.1 : FIELD_NAME
             | quoted_field_name

quoted_field_name.1 : QUOTED_FIELD_NAME



?literal.2 : KEYWORD_LITERAL -> keyword_literal
           | DATE            -> date
           | DATETIME        -> datetime
           | TIME            -> time
           | ESCAPED_STRING  -> string
           | SIGNED_NUMBER   -> number
           | list
           | empty_list

DATE.2 : INT "-" INT "-" INT
TIME.2 : INT ":" INT (":" INT ("." INT)?)?
DATETIME.2 : DATE "T" TIME

KEYWORD_LITERAL : "NuLL"i
                | "TRUE"i
                | "FaLSE"i

FIELD_NAME : ("_"|LETTER) ("_"|LETTER|DIGIT)*
QUOTED_FIELD_NAME : "{" /[^}]/* "}"

empty_list : "[]"
list : "[" literal ("," literal)* "]"

_STRING_INNER: /(.|\\n)*?/
_STRING_ESC_INNER: _STRING_INNER /(?<!\\\\)(\\\\\\\\)*?/
ESCAPED_STRING : "\\"" _STRING_ESC_INNER "\\""



%import common.INT
%import common.SIGNED_NUMBER
%import common.LETTER
%import common.DIGIT

%import common.WS
%ignore WS
"""

# The instance of the grammar parser is created only once
# then stored in _grammar_parser for later reuse
_grammar_parser = None


def filter_parser():
    """
    :return: A singleton instance of Lark grammar parser for filter expression
    """
    global _grammar_parser
    if _grammar_parser is None:
        _grammar_parser = Lark(filter_grammar, parser="lalr")
    return _grammar_parser


def literal_parser():
    """
    This is used to test literals parsing

    :return: An instance of Lark grammar parser for parsing only a literal
       value (int, string, list, date, etc.) from a filter expression. This
       is used for testing the parsing of these literals.
    """
    return Lark(filter_grammar, parser="lalr", start="literal")


def _list_to_sql(value):
    s = ",".join(f'"{x}"' if isinstance(x, str) else to_sql(x) for x in value)
    return f"'[{s}]'"


_type_to_sql = {
    type(None): lambda x: "NULL",
    str: lambda x: f"'{x}'",
    int: str,
    float: str,
    # datetime.time: lambda x: f"'{x.isoformat()}'",
    # datetime.datetime: lambda x: f"'{x.isoformat()}'",
    # datetime.date: lambda x: f"'{x.isoformat()}'",
    bool: lambda x: ("1" if x else "0"),
    list: _list_to_sql,
}


def to_sql(value):
    global _type_to_sql

    return _type_to_sql[type(value)](value)


class Field(str):
    pass


class FilterToSQL(Transformer):
    """
    Instance of this class are passed to Lark parser when parsing a document
    selection filter string in order to create an object that can be used to
    select items from the database. FilterToSQL implements methods that are
    common to all engines and does not produce anything because the query
    objectis specific to each engine. Therefore, engine class must use a
    subclass of FilterToSQL that implements the following methods:

        build_condition_all
        build_condition_literal_in_list_field
        build_condition_field_in_list_field
        build_condition_field_in_list
        build_condition_field_op_field
        build_condition_value_op_field
        build_condition_negation
        build_condition_combine_conditions
    """

    keyword_literals = {
        "true": True,
        "false": False,
        "null": None,
    }

    sql_operators = {
        "==": "IS",
        "!=": "IS NOT",
        "ilike": "LIKE",
    }

    no_list_operators = {">", "<", ">=", "<=", "like", "ilike"}

    def __init__(self, dbcollection):
        self.dbcollection = dbcollection

    def all(self, items):
        return self.build_condition_all()

    def condition(self, items):
        left_operand, operator, right_operand = items
        operator_str = str(operator).lower()
        if operator_str == "in":
            if isinstance(right_operand, Field):
                if left_operand is None or isinstance(
                    left_operand,
                    (
                        str
                        | int
                        | float
                        | bool
                        | datetime.date
                        | datetime.time
                        | datetime.datetime
                    ),
                ):
                    return self.build_condition_literal_in_list_field(
                        left_operand, right_operand
                    )
                elif isinstance(left_operand, Field):
                    return self.build_condition_field_in_list_field(
                        left_operand, right_operand
                    )
                else:
                    raise ValueError(
                        "Left operand of IN <list field> must be a "
                        f'field, string, number, boolean, date, time or null but "{left_operand}" '
                        "was used"
                    )
            elif isinstance(right_operand, list):
                if isinstance(left_operand, Field):
                    return self.build_condition_field_in_list(
                        left_operand, right_operand
                    )
                else:
                    raise ValueError(
                        "Left operand of IN <list> must be a "
                        f'simple field but "{left_operand}" was used'
                    )
            else:
                raise ValueError(
                    "Right operand of IN must be a list or a "
                    f'list field but "{right_operand}" was used'
                )

        if isinstance(left_operand, Field):
            if isinstance(right_operand, Field):
                return self.build_condition_field_op_field(
                    left_operand, operator_str, right_operand
                )
            else:
                return self.build_condition_field_op_value(
                    left_operand, operator_str, right_operand
                )
        else:
            if isinstance(right_operand, Field):
                return self.build_condition_value_op_field(
                    left_operand, operator_str, right_operand
                )
            else:
                raise ValueError(
                    "Either left or right operand of a condition must be a field name"
                )

    def negation(self, items):
        return self.build_condition_negation(items[0])

    def conditions(self, items):
        stack = list(items)
        result = stack.pop(0)
        while stack:
            operator_str = stack.pop(0).lower()
            right_operand = stack.pop(0)
            if result is None:
                if operator_str == "and":
                    result = right_operand
                    continue
                elif operator_str == "or":
                    result = None
                    continue
                left_operand = ["1"]
            else:
                left_operand = result
            if right_operand is None:
                if operator_str == "and":
                    result = left_operand
                    continue
                elif operator_str == "or":
                    result = None
                    continue
                right_operand = ["1"]
            result = self.build_condition_combine_conditions(
                result, operator_str, right_operand
            )
        return result

    def string(self, items):
        return ast.literal_eval(items[0].replace("\n", "\\n"))

    def number(self, items):
        return float(items[0])

    def date(self, items):
        return dateutil.parser.parse(items[0]).date()

    def time(self, items):
        return dateutil.parser.parse(items[0]).time()

    def datetime(self, items):
        return dateutil.parser.parse(items[0])

    def keyword_literal(self, items):
        return self.keyword_literals[items[0].lower()]

    def empty_list(self, items):
        return []

    def list(self, items):
        return items

    def field_name(self, items):
        field = items[0]
        # Checks for literal due to a bug in Lark
        literal = self.keyword_literals.get(field.lower(), self)
        if literal is not self:
            return literal
        if field in self.dbcollection.fields:
            return Field(f"[{field}]")
        elif self.dbcollection.catchall_column:
            return Field(
                f"json_extract([{self.dbcollection.catchall_column}],'$.\"{field}\"')"
            )
        else:
            raise ValueError(
                f'Filter uses unknown field "{field}" in collection "{self.dbcollection.name}" that does not support it'
            )

    def quoted_field_name(self, items):
        return Field(items[0][1:-1])

    def build_condition_all(self):
        """
        Return a selection query that select all documents. This query
        is directly given to the engine and never combined with other
        queries.
        """
        return None

    def build_condition_literal_in_list_field(self, value, list_field):
        """
        Builds a condition checking if a constant value is in a list field

        :param value: Python literal

        :param list_field: field object as returned by Database.get_field

        """
        return [
            f"{list_field} IS NOT NULL AND "
            f"{to_sql(value)} IN (SELECT value FROM json_each({list_field}))"
        ]

    def build_condition_field_in_list_field(self, field, list_field):
        """
        Builds a condition checking if a field value is in another
        list field value

        :param field: field object as returned by Database.get_field
        :param list_field: field object as returned by Database.get_field
        """
        return [
            f"{list_field} IS NOT NULL AND "
            f"{field} IN (SELECT value FROM json_each({list_field}))"
        ]

    def build_condition_field_in_list(self, field, list_value):
        """
        Builds a condition checking if a field value is a
        constant list value

        :param field: field object as returned by Database.get_field
        :param list_value: Python list containing literals
        """
        if None in list_value:
            list_value.remove(None)
            where = [f"{field} IS NULL OR {field} "]
        else:
            where = [f"{field} "]
        if not list_value:
            return ["0"]
        elif len(list_value) == 1:
            where.append(f"IS {to_sql(list_value[0])}")
        else:
            where.append(f"IN ({','.join(to_sql(i) for i in list_value)})")
        return where

    def build_condition_field_op_field(self, left_field, operator_str, right_field):
        """
        Builds a condition comparing the content of two fields with an operator.

        :param left_field: field object as returned by Database.get_field
        :param operator: string containing one of the CONDITION_OPERATOR
                         defined in the grammar (in lowercase)
        :param right_field: field object as returned by Database.get_field
        """
        sql_operator = self.sql_operators.get(operator_str, operator_str)
        if operator_str == "ilike":
            return [f"UPPER({left_field}) {sql_operator} UPPER({right_field})"]
        else:
            return [f"{left_field} {sql_operator} {right_field}"]

    def build_condition_field_op_value(self, field, operator_str, value):
        """
        Builds a condition comparing the content of a field with a constant
        value using an operator.

        :param field: field object as returned by Database.get_field
        :param operator_str: string containing one of the CONDITION_OPERATOR
                             defined in the grammar (in lowercase)
        :param value: Python value (None, string number, boolean or date/time)
        """
        if isinstance(value, list):
            if operator_str in self.no_list_operators:
                raise ValueError(
                    f"operator {operator_str} cannot be used with value of list type"
                )
        if operator_str == "ilike":
            field = f"UPPER({field})"
            if isinstance(value, str):
                value = value.upper()
        else:
            field = f"{field}"
        sql_operator = self.sql_operators.get(operator_str, operator_str)
        return [f"{field} {sql_operator} {to_sql(value)}"]

    def build_condition_value_op_field(self, value, operator_str, field):
        """
        Builds a condition comparing a constant value with the content of a
        field withusing an operator.

        :param value: Python value (None, string number, boolean or date/time)
        :param operator_str: string containing one of the CONDITION_OPERATOR
                             defined in the grammar (in lowercase)
        :param field: field object as returned by Database.get_field
        """
        if isinstance(value, list):
            if operator_str in self.no_list_operators:
                raise ValueError(
                    f"operator {operator_str} cannot be used with value of list type"
                )
        if operator_str == "ilike":
            field = f"UPPER({field})"
            if isinstance(value, str):
                value = value.upper()
        sql_operator = self.sql_operators.get(operator_str, operator_str)
        return [f"{to_sql(value)} {sql_operator} {field}"]

    def build_condition_negation(self, condition):
        """
        Builds a condition inverting another condition.

        :param condition: condition object returned by one of the
                          build_condition_*() method (except
                          build_condition_all)
        """
        if condition is None:
            return ["0"]
        return ["NOT", "("] + condition + [")"]

    def build_condition_combine_conditions(
        self, left_condition, operator_str, right_condition
    ):
        """
        Builds a condition that combines two conditions with an operator.

        :param left_condition: condition object returned by one of the
                               build_condition_*() method (except
                               build_condition_all)
        :param operator_str: string containing one of the BOOLEAN_OPERATOR
                             defined in the grammar (in lowercase)
        :param right_condition: condition object returned by one of the
                                build_condition_*() method (except
                                build_condition_all)
        """
        return (
            ["("] + left_condition + [")", operator_str, "("] + right_condition + [")"]
        )



import importlib
import os
import re
import sqlite3
import typing
from uuid import uuid4

import requests
from cryptography.fernet import Fernet, InvalidToken

import populse_db.storage
from populse_db.database import json_decode, json_encode, str_to_type

from . import Database
from .database import populse_db_table


def serialize_exception(e):
    try:
        # Import tblib in this function avoid to make it a mandatory dependency
        import tblib

        tb = tblib.Traceback(e.__traceback__)
        result = {
            "class_module": e.__class__.__module__,
            "class_name": e.__class__.__name__,
            "traceback": tb.to_dict(),
        }

        kwargs = e.__getstate__() if hasattr(e, "__getstate__") else None
        if kwargs is None:
            result["args"] = e.args
        else:
            kwargs.pop("lineno", None)
            kwargs.pop("colno", None)
            result["kwargs"] = kwargs
    except Exception as e2:
        result = {
            "class_module": e2.__class__.__module__,
            "class_name": e2.__class__.__name__,
            "args": [f"Error while managing exception ({e}): {e2}"],
        }

    return result


def deserialize_exception(je):
    exception_class = getattr(
        importlib.import_module(je["class_module"]), je["class_name"]
    )
    exception = exception_class(*je.get("args", []), **je.get("kwargs", {}))
    tb = je.get("traceback")
    if tb is not None:
        # Import tblib in this function avoid to make it a mandatory dependency
        import tblib

        tb = tblib.Traceback.from_dict(tb)
        exception.with_traceback(tb.as_traceback())
    return exception


class StorageAPI:
    """
    Select between StorageFileAPI() and StorageServerAPI()
    """

    def __new__(
        cls,
        database_file: str,
        timeout: float | None = None,
        create: bool = False,
        echo_sql: typing.TextIO | None = None,
    ):
        if re.match("^https?:.*", database_file):
            return StorageServerAPI(database_file)
        if os.path.exists(database_file) or create:
            cnx = sqlite3.connect(database_file, timeout=10)
            # Check if storage_server table exists
            cur = cnx.execute(
                "SELECT COUNT(*) FROM sqlite_master "
                f"WHERE type='table' AND name='{populse_db_table}'"
            )
            if cur.fetchone()[0]:
                row = cnx.execute(
                    f"SELECT _json FROM [{populse_db_table}] WHERE category='server' AND key='url'"
                ).fetchone()
                if row:
                    return StorageServerAPI(row[0])
        return StorageFileAPI(database_file, timeout, create, echo_sql)


class StorageFileAPI:
    _read_only_error = "database is read-only"

    def __init__(
        self,
        database_file: str,
        timeout: float | None = None,
        create: bool = False,
        echo_sql: typing.TextIO | None = None,
    ):
        self.key = Fernet.generate_key()
        self.database = Database(
            database_file, timeout=timeout, create=create, echo_sql=echo_sql
        )
        self.sessions = {}

    @staticmethod
    def _init_database(dbs):
        # Initialize schema
        if not dbs.has_collection(populse_db.storage.Storage.default_collection):
            dbs.add_collection(
                populse_db.storage.Storage.default_collection,
                populse_db.Storage.default_field,
            )
            dbs[populse_db.storage.Storage.default_collection][
                populse_db.storage.Storage.default_document_id
            ] = {}

    @staticmethod
    def _parse_path(dbs, path):
        if not path:
            return (None, None, None, None)
        if dbs.has_collection(path[0]):
            collection = dbs[path[0]]
            path = path[1:]
        else:
            collection = dbs[populse_db.storage.Storage.default_collection]
        if populse_db.storage.Storage.default_field in collection.primary_key:
            document_id = populse_db.storage.Storage.default_document_id
        else:
            if not path:
                return (collection, None, None, None)
            document_id = path[0]
            path = path[1:]
        if not path:
            return (collection, document_id, None, None)
        field = path[0]
        path = path[1:]
        return (collection, document_id, field, path)

    def access_token(self):
        # For local file, we chose to not check access rights
        # at this time but to grant write access and let the
        # filesystem raise an exception when a write access is
        # tried on a read only file.
        f = Fernet(self.key)
        return f.encrypt(b"write").decode()

    def connect(self, access_token, exclusive, write, create):
        connection_id = str(uuid4())
        f = Fernet(self.key)
        try:
            access_rights = f.decrypt(access_token.encode()).decode()
        except InvalidToken:
            raise PermissionError("invalid token") from None
        if not write and access_rights in ("read", "write"):
            dbs = self.database.session(exclusive=exclusive, create=False)
        elif access_rights == "write":
            # Write access must be in an exclusive transaction to avoid
            # "Database is locked" error in SQLite with parallel accesses:
            # https://www2.sqlite.org/cvstrac/wiki?p=DatabaseIsLocked
            dbs = self.database.session(exclusive=True, create=create)
        else:
            raise PermissionError("database access refused")
        if dbs:
            self.sessions[connection_id] = (dbs, write)
            if write:
                self._init_database(dbs)
            return connection_id
        return None

    def _get_database_session(self, connection_id, write):
        session = self.sessions.get(connection_id)
        if session is None:
            raise ValueError("invalid storage connection id")
        dbs, writable = session
        if write and not writable:
            raise PermissionError(self._read_only_error)
        return dbs

    def add_schema_collections(self, connection_id, schema_to_collections):
        dbs = self._get_database_session(connection_id, write=True)
        if not dbs.has_collection(populse_db.storage.Storage.schema_collection):
            dbs.add_collection(populse_db.storage.Storage.schema_collection, "name")
            schema_collection = dbs[populse_db.storage.Storage.schema_collection]
            schema_collection.add_field("version", str)
            schema_collection[schema_to_collections["name"]] = {
                "version": schema_to_collections["version"]
            }
        collections = schema_to_collections["collections"]
        for collection_name, fields_definition in collections.items():
            if not dbs.has_collection(collection_name):
                primary_keys = {}
                for field_name, field_def in fields_definition.items():
                    type_str, kwargs = field_def
                    if kwargs.get("primary_key", False):
                        primary_keys[field_name] = str_to_type(type_str)
                dbs.add_collection(collection_name, primary_keys)
                if populse_db.storage.Storage.default_field in primary_keys:
                    dbs[collection_name][
                        populse_db.storage.Storage.default_document_id
                    ] = {}

            collection = dbs[collection_name]
            for field_name, field_def in fields_definition.items():
                type_str, kwargs = field_def
                field = collection.fields.get(field_name)
                if field:
                    type_python = str_to_type(type_str)
                    if field["type"] != type_python:
                        raise ValueError(
                            f"database has a {collection_name}.{field_name} field "
                            f"with type {field['type']} but schema requires type {type_python}"
                        )
                    if field["primary_key"] != kwargs.get("primary_key", False):
                        raise ValueError(
                            f"primary key difference between database and schema for {collection_name}.{field_name}"
                        )
                else:
                    collection.add_field(field_name, str_to_type(type_str), **kwargs)

    def add_collection(self, connection_id, name, primary_key):
        dbs = self._get_database_session(connection_id, write=True)
        collection = dbs.get_collection(name)
        if collection is None:
            dbs.add_collection(name, primary_key)
        else:
            # Check that primary key of existing collection is compatible
            # with requested one.
            if isinstance(primary_key, str):
                dict_primary_key = {primary_key: str}
            elif not isinstance(primary_key, dict):
                dict_primary_key = dict((i, str) for i in primary_key)
            else:
                dict_primary_key = dict(
                    (k, str_to_type(v)) for k, v in primary_key.items()
                )
            if collection.primary_key != dict_primary_key:
                raise ValueError(
                    f"primary key {primary_key} is not compatible with the one defined for collection {name}"
                )

    def add_field(
        self,
        connection_id,
        collection_name,
        field_name,
        field_type,
        description=None,
        index=False,
    ):
        dbs = self._get_database_session(connection_id, write=True)
        if collection_name is None:
            collection_name = populse_db.storage.Storage.default_collection
        collection = dbs.get_collection(collection_name)
        if collection is None:
            raise ValueError(f'No collection named "{collection_name}"')
        field = collection.fields.get(field_name)
        if field is None:
            collection.add_field(
                field_name, field_type, description=description, index=index
            )
        else:
            # Check compatibility with existing field
            if field["type"] != str_to_type(field_type):
                raise ValueError(
                    f"Incompatible type for field {field_name} of collection {collection_name}, requested {field_type} but existing database has {field['type']}"
                )

    def remove_field(self, connection_id, collection_name, field_name):
        """
        Removes a specified field from a collection in the database.

        Args:
            connection_id (str): The identifier for the database connection.
            collection_name (str): The name of the collection.
                                   Defaults to the storage's default
                                   collection.
            field_name (str): The name of the field to remove.

        Raises:
            ValueError: If the specified collection does not exist.
        """
        # Retrieve the database session with write access
        dbs = self._get_database_session(connection_id, write=True)
        # Use default collection name if none is provided
        collection_name = (
            collection_name or populse_db.storage.Storage.default_collection
        )
        # Retrieve the specified collection
        collection = dbs.get_collection(collection_name)

        if collection is None:
            raise ValueError(f'No collection named "{collection_name}"')

        # Remove the field if it exists
        if collection.fields.get(field_name):
            collection.remove_field(field_name)

    def disconnect(self, connection_id, rollback):
        dbs = self._get_database_session(connection_id, write=False)
        dbs.close(rollback)

    def get(
        self,
        connection_id,
        path,
        default=None,
        fields=None,
        as_list=None,
        distinct=False,
    ):
        dbs = self._get_database_session(connection_id, write=False)
        collection, document_id, field, path = self._parse_path(dbs, path)
        if not collection:
            raise ValueError("cannot get the whole content of a database")
        if distinct and document_id:
            raise ValueError("distinct is only allowed on collection")
        if field:
            if fields:
                raise ValueError(
                    "fields selection is only allowed for collection or document"
                )
            if as_list:
                raise ValueError("as_list is only allowed for collection or document")
            row = collection.document(document_id, fields=[field], as_list=True)
            if not row:
                return default
            value = row[0]
            if value is None:
                return default
            for i in path:
                if isinstance(value, list) and isinstance(i, str):
                    i = int(i)
                try:
                    value = value[i]
                except (KeyError, IndexError):
                    return default
            return value
        elif document_id:
            document = collection.document(document_id, fields=fields, as_list=as_list)
            if not document:
                return default
            if document_id == populse_db.storage.Storage.default_document_id:
                del document[document_id]
            return document
        else:
            return list(
                collection.documents(fields=fields, as_list=as_list, distinct=distinct)
            )

    def count(self, connection_id, path, query=None):
        dbs = self._get_database_session(connection_id, write=False)
        collection, document_id, field, path = self._parse_path(dbs, path)
        if path or field:
            raise ValueError("only collections can be counted")
        if document_id:
            document_id = collection.document_id(document_id)
            primary_key = list(collection.primary_key)
            document_query = " and ".join(
                f'{{{primary_key[i]}}}=="{document_id[i]}"'
                for i in range(len(document_id))
            )
            if query:
                query = f"{query} and {document_query}"
            else:
                query = document_query
        return collection.count(query)

    def primary_key(self, connection_id, path):
        dbs = self._get_database_session(connection_id, write=False)
        collection, document_id, field, path = self._parse_path(dbs, path)
        if not collection or document_id:
            raise ValueError("primary_key is only allowed on collections")
        return list(collection.primary_key)

    def set(self, connection_id, path, value):
        dbs = self._get_database_session(connection_id, write=True)
        collection, document_id, field, path = self._parse_path(dbs, path)
        if not collection:
            raise ValueError("cannot set the whole content of a database")
        if field:
            if path:
                db_value = collection.document(
                    document_id, fields=[field], as_list=True
                )[0]
                container = db_value
                for i in path[:-1]:
                    if isinstance(container, list) and isinstance(i, str):
                        i = int(i)
                    container = container[i]
                i = path[-1]
                if isinstance(container, list) and isinstance(i, str):
                    i = int(i)
                container[i] = value
                value = db_value
            collection.update_document(document_id, {field: value})
        elif document_id:
            collection[document_id] = value
        else:
            collection.delete(None)
            for document in value:
                collection.add(document)

    def delete(self, connection_id, path):
        dbs = self._get_database_session(connection_id, write=True)
        collection, document_id, field, path = self._parse_path(dbs, path)
        if not collection:
            raise ValueError("cannot delete the whole content of a database")
        if not document_id:
            raise ValueError("cannot delete a collection in a data session")
        if field:
            if path:
                db_value = collection.document(
                    document_id, fields=[field], as_list=True
                )[0]
                container = db_value
                for i in path[:-1]:
                    container = container[i]
                del container[path[-1]]
            else:
                db_value = None
            collection.update_document(document_id, {field: db_value})
        else:
            del collection[document_id]

    def update(self, connection_id, path, value):
        dbs = self._get_database_session(connection_id, write=True)
        collection, document_id, field, path = self._parse_path(dbs, path)
        if not document_id:
            raise ValueError("cannot update a database or a collection")
        if field:
            db_value = collection.document(document_id, fields=[field], as_list=True)[0]
            if path:
                container = db_value
                for i in path:
                    container = container[i]
                container.update(value)
            else:
                db_value.update(value)
            collection.update_document(document_id, {field: db_value})
        else:
            collection.update_document(document_id, value)

    def append(self, connection_id, path, value):
        dbs = self._get_database_session(connection_id, write=True)
        collection, document_id, field, path = self._parse_path(dbs, path)
        if not collection:
            raise ValueError("cannot set the whole content of a database")
        if field:
            db_value = collection.document(document_id, fields=[field], as_list=True)[0]
            container = db_value
            for i in path:
                container = container[i]
            container.append(value)
            collection.update_document(document_id, {field: db_value})
        elif document_id:
            raise TypeError(
                f"cannot append to document {document_id} in collection {collection.name}"
            )
        else:
            collection.add(value)

    def search(
        self, connection_id, path, query, fields=None, as_list=None, distinct=False
    ):
        dbs = self._get_database_session(connection_id, write=False)
        collection, document_id, field, path = self._parse_path(dbs, path)
        if path or field:
            raise ValueError("only collections can be searched")
        if document_id:
            document_id = collection.document_id(document_id)
            primary_key = list(collection.primary_key)
            document_query = " and ".join(
                f'{{{primary_key[i]}}}=="{document_id[i]}"'
                for i in range(len(document_id))
            )
            if query:
                query = f"{query} and {document_query}"
            else:
                query = document_query
        result = list(
            collection.filter(query, fields=fields, as_list=as_list, distinct=distinct)
        )
        return result

    def search_and_delete(self, connection_id, path, query):
        dbs = self._get_database_session(connection_id, write=False)
        collection, document_id, field, path = self._parse_path(dbs, path)
        if path or field:
            raise ValueError("only collections can be searched")
        if document_id:
            document_id = collection.document_id(document_id)
            primary_key = list(collection.primary_key)
            document_query = " and ".join(
                f'{{{primary_key[i]}}}=="{document_id[i]}"'
                for i in range(len(document_id))
            )
            if query:
                query = f"{query} and {document_query}"
            else:
                query = document_query
        collection.delete(query)

    def distinct_values(self, connection_id, path, field):
        dbs = self._get_database_session(connection_id, write=False)
        collection, document_id, f, path = self._parse_path(dbs, path)
        if path or f or document_id:
            raise ValueError("only collections support distinct values searching")
        for row in collection.documents(fields=[field], as_list=True, distinct=True):
            yield row[0]

    def clear_database(self, connection_id, path):
        dbs = self._get_database_session(connection_id, write=True)
        collection, document_id, f, path = self._parse_path(dbs, path)
        if collection:
            raise ValueError("clear_database can only be called on a whole database")
        dbs.clear()
        self._init_database()

    def has_collection(self, connection_id, path, collection):
        dbs = self._get_database_session(connection_id, write=False)
        c, document_id, f, path = self._parse_path(dbs, path)
        if c:
            raise ValueError("has_collection can only be called on a whole database")
        return dbs.has_collection(collection)

    def collection_names(self, connection_id, path):
        dbs = self._get_database_session(connection_id, write=False)
        collection, document_id, f, path = self._parse_path(dbs, path)
        if collection:
            raise ValueError("collection_names can only be called on a whole database")
        return [
            i.name
            for i in dbs.get_collections()
            if i.name
            not in {
                populse_db.storage.Storage.default_collection,
                populse_db.storage.Storage.schema_collection,
            }
        ]

    def keys(self, connection_id, path):
        dbs = self._get_database_session(connection_id, write=False)
        collection, document_id, f, path = self._parse_path(dbs, path)
        if not collection or path or f or document_id:
            raise ValueError("only collections support keys()")
        return collection.fields.keys()


class StorageServerAPI:
    def __init__(self, url):
        self.url = url

    def access_token(self):
        return self._call("get", "access_token", None)

    def _call(self, method, route, payload, decode=False):
        response = requests.request(
            method,
            f"{self.url}/{route}",
            json=payload,
        )
        if response.status_code == 500:
            exc = deserialize_exception(response.json())
            raise exc
        response.raise_for_status()
        result = response.json()
        if decode:
            result = json_decode(result)
        return result

    def connect(self, access_token, exclusive, write, create):
        return self._call(
            "post",
            "connection",
            dict(
                access_token=access_token,
                exclusive=bool(exclusive),
                write=bool(write),
                create=bool(create),
            ),
        )

    def disconnect(self, connection_id, rollback):
        return self._call(
            "delete",
            "connection",
            dict(connection_id=connection_id, rollback=rollback),
        )

    def add_schema_collections(self, connection_id, schema_to_collections):
        return self._call(
            "post",
            "schema_collection",
            dict(
                connection_id=connection_id,
                schema_to_collections=schema_to_collections,
            ),
        )

    def add_collection(self, connection_id, name, primary_key):
        return self._call(
            "post",
            f"schema/{name}",
            dict(connection_id=connection_id, primary_key=primary_key),
        )

    def add_field(
        self,
        connection_id,
        collection_name,
        field_name,
        field_type,
        description=None,
        index=False,
    ):
        return self._call(
            "post",
            f"schema/{collection_name}/{field_name}",
            dict(
                connection_id=connection_id,
                field_type=field_type,
                description=description,
                index=index,
            ),
        )

    def remove_field(
        self,
        connection_id,
        collection_name,
        field_name,
    ):
        return self._call(
            "delete",
            f"schema/{collection_name}/{field_name}",
            dict(
                connection_id=connection_id,
            ),
        )

    def get(
        self,
        connection_id,
        path,
        default=None,
        fields=None,
        as_list=None,
        distinct=False,
    ):
        return self._call(
            "get",
            "data",
            dict(
                connection_id=connection_id,
                path=path,
                default=default,
                fields=fields,
                as_list=as_list,
                distinct=distinct,
            ),
            decode=True,
        )

    def count(self, connection_id, path, query=None):
        return self._call(
            "get",
            "count",
            dict(connection_id=connection_id, path=path, query=query),
        )

    def primary_key(self, connection_id, path):
        return self._call(
            "get",
            "primary_key",
            dict(connection_id=connection_id, path=path),
        )

    def set(self, connection_id, path, value):
        return self._call(
            "post",
            "data",
            dict(connection_id=connection_id, path=path, value=json_encode(value)),
        )

    def delete(self, connection_id, path):
        return self._call(
            "delete", "data", dict(connection_id=connection_id, path=path)
        )

    def update(self, connection_id, path, value):
        return self._call(
            "put",
            "data",
            dict(connection_id=connection_id, path=path, value=json_encode(value)),
        )

    def append(self, connection_id, path, value):
        return self._call(
            "patch",
            "data",
            dict(connection_id=connection_id, path=path, value=json_encode(value)),
        )

    def search(
        self, connection_id, path, query, fields=None, as_list=None, distinct=False
    ):
        return self._call(
            "get",
            "search",
            dict(
                connection_id=connection_id,
                path=path,
                query=query,
                fields=fields,
                as_list=as_list,
                distinct=distinct,
            ),
            decode=True,
        )

    def search_and_delete(self, connection_id, path, query):
        return self._call(
            "delete",
            "search",
            dict(connection_id=connection_id, path=path, query=query),
            decode=True,
        )

    def distinct_values(self, connection_id, path, field):
        return self._call(
            "get",
            "distinct",
            dict(connection_id=connection_id, path=path, field=field),
            decode=True,
        )

    def clear_database(self, connection_id, path):
        return self._call(
            "delete",
            "",
            dict(connection_id=connection_id, path=path),
        )

    def has_collection(self, connection_id, path, collection):
        return self._call(
            "get",
            "has_collection",
            dict(connection_id=connection_id, path=path, collection=collection),
        )

    def collection_names(self, connection_id, path):
        return self._call(
            "get",
            "collection_names",
            dict(connection_id=connection_id, path=path),
        )

    def keys(self, connection_id, path):
        return self._call(
            "get",
            "keys",
            dict(connection_id=connection_id, path=path),
        )



from datetime import datetime

schemas = [
    {
        "version": "1.0.0",
        "schema": {
            # A collection of executions to track data provenance
            "execution": [
                {
                    "execution_id": ["str", {"primary_key": True}],
                    "start_time": "datetime",
                    "end_time": datetime,
                    "status": str,
                    "capsul_executable": str,
                    "capsul_parameters": dict,
                    "software": str,
                    "software_module": str,
                    "software_version": str,
                }
            ],
        },
    }
]



import json
from datetime import date, datetime, time

import dateutil

populse_db_table = "populse_db"


def check_value_type(value, field_type):
    """
    Checks the type of a value

    :param value: value to check

    :param field_type: type that the value is supposed to have

    :return: true if the value is ``None`` or if the value is of that type,
        ``False`` otherwise
    """

    if field_type is None:
        return False
    if value is None:
        return True
    origin = getattr(field_type, "__origin__", None)
    if origin:
        # field_type is a parameterized type such as list[str]
        # origin is the parent type (e.g. list)
        if isinstance(value, origin):
            # The following code works only on list[...]
            # because other parameterized types are not
            # supported.
            item_type = field_type.__args__[0]
            for v in value:
                if not check_value_type(v, item_type):
                    return False
            return True
    elif field_type is float:
        return isinstance(value, int | float)
    else:
        return isinstance(value, field_type)
    return False


def type_to_str(type):
    """Convert a Python type to a string.

    Examples:

        - ``type_to_str(str) == 'str'``
        - ``type_to_str(list[str]) == 'list[str]'``
    """
    args = getattr(type, "__args__", None)
    if args:
        return f"{type.__name__}[{','.join(type_to_str(i) for i in args)}]"
    else:
        return type.__name__


_type_to_sqlite = {
    str: "text",
}


def type_to_sqlite(type):
    """
    Like type_to_str(type) but for internal use in SQLite column type
    definitions in order to avoid conversion problems due to SQlite type
    affinity. See https://www.sqlite.org/datatype3.html
    """
    result = _type_to_sqlite.get(type)
    if result is None:
        args = getattr(type, "__args__", None)
        if args:
            result = f"{type.__name__}[{','.join(type_to_sqlite(i) for i in args)}]"
        else:
            result = type.__name__
    return result


_str_to_type = {
    type_to_sqlite(i): i
    for i in (str, int, float, bool, date, datetime, time, dict, list)
}
_str_to_type.update(
    {
        type_to_str(i): i
        for i in (str, int, float, bool, date, datetime, time, dict, list)
    }
)


def str_to_type(str):
    """Convert a string to a Python type.

    Examples:

        - ``str_to_type('str') == str``
        - ``str_to_type('list[str]') == list[str]``
    """
    global _str_to_type

    if not str:
        return None
    s = str.split("[", 1)
    if len(s) == 1:
        result = _str_to_type.get(s[0])
    else:
        args = tuple(str_to_type(i) for i in s[1][:-1].split(","))
        result = _str_to_type.get(s[0])
        if result:
            result = result[args]
    if not result:
        raise ValueError(f'invalid type: "{str}"')
    return result


def python_value_type(value):
    """
    Returns the Python type corresponding to a Python value.
    This type can be used in add_field(s) method.
    For list values, only the first item is considered to get the subtype
    of the list.

    Examples:

        - ``python_value_type('a value') == str``
        - ``python_value_type([]) == list``
        - ``python_value_type([1, 2, 3]) == list[int]``
        - ``python_value_type(['string', 2, {}]) == list[str]``
        - ``python_value_type({'one': 1, 'two': 2}) == dict``
    """
    if isinstance(value, list) and value:
        return list[type(value[0])]
    type(value)


class DatabaseSession:
    """
    Base class

    methods:

        *Database related methods:*

        - :py:meth:`execute`
        - :py:meth:`commit`
        - :py:meth:`rollback`

        *Database configuration methods:*

        - :any:`settings`
        - :py:meth:`set_settings`

        *Collections related methods*

        - :py:meth:`add_collection`
        - :py:meth:`remove_collection`
        - :py:meth:`has_collection`
        - :py:meth:`__getitem__`
        - :py:meth:`collections`

        *Obsolete methods kept for backward compatibility*

        - :py:meth:`get_collection`
        - :py:meth:`get_collections`
        - :py:meth:`get_collections_names`
        - :py:meth:`add_field`
        - :py:meth:`remove_field`
        - :py:meth:`get_field`
        - :py:meth:`get_fields_names`
        - :py:meth:`get_fields`
        - :py:meth:`set_values`
        - :py:meth:`has_document`
        - :py:meth:`get_document`
        - :py:meth:`get_documents_ids`
        - :py:meth:`get_documents`
        - :py:meth:`remove_document`
        - :py:meth:`add_document`
        - :py:meth:`filter_documents`

    .. automethod:: __getitem__
    """

    default_primary_key = "primary_key"

    def execute(self, *args, **kwargs):
        raise NotImplementedError()

    def commit(self):
        raise NotImplementedError()

    def rollback(self):
        raise NotImplementedError()

    def settings(self, category, key, default=None):
        raise NotImplementedError()

    def set_settings(self, category, key, value):
        raise NotImplementedError()

    def add_collection(self, name, primary_key=default_primary_key):
        """
        Adds a collection

        :param name: New collection name (str, must not be existing)

        :param primary_key: New collection primary_key column (str) => "index" by default

        :raise ValueError: - If the collection is already existing
                           - If the collection name is invalid
                           - If the primary_key is invalid
        """
        raise NotImplementedError()

    def remove_collection(self, name):
        """
        Removes a collection

        :param name: Collection to remove (str, must be existing)

        :raise ValueError: If the collection does not exist
        """
        raise NotImplementedError()

    def has_collection(self, name):
        """
        Check if a collection with the given name exists.
        """
        raise NotImplementedError()

    def __getitem__(self, collection_name):
        """Return a collection object given its name."""
        raise NotImplementedError()

    def collections(self):
        """
        Iterates over collections

        :return: generator
        """

        yield from self

    def get_collection(self, name):
        """
        .. deprecated:: 3.0
            Use ``db_session[name]`` instead
        """
        try:
            return self[name]
        except ValueError:
            return None

    def get_collections(self):
        """
        .. deprecated:: 3.0
            Use :py:meth:`collections()` instead
        """
        return self.collections()

    def get_collections_names(self):
        """
        .. deprecated:: 3.0
            Use ``(i.name for i in db_session)`` instead
        """
        return (i.name for i in self)

    def add_field(self, collection, name, field_type, description=None, index=False):
        """
        .. deprecated:: 3.0
            Use ``db_session[collection].add_field(...)`` instead.
            See :py:meth:`DatabaseCollection.add_field`.
        """
        self[collection].add_field(
            name, field_type, description=description, index=index
        )

    def remove_field(self, collection, field):
        """
        .. deprecated:: 3.0
            Use ``db_session[collection].remove_field(...)`` instead.
            See :py:meth:`DatabaseCollection.remove_field`.
        """
        self[collection].remove_field(field)

    def get_field(self, collection, name):
        """
        .. deprecated:: 3.0
            Use ``db_session[collection].fields.get(name)`` instead.
            See :py:attr:`DatabaseCollection.fields`.
        """
        try:
            return self[collection].fields.get(name)
        except ValueError:
            return None

    def get_fields_names(self, collection):
        """
        .. deprecated:: 3.0
            Use ``db_session[collection].fields.keys()`` instead.
            See :py:attr:`DatabaseCollection.fields`.
        """
        try:
            return self[collection].fields.keys()
        except ValueError:
            return ()

    def get_fields(self, collection):
        """
        .. deprecated:: 3.0
            Use ``db_session[collection].fields.values()`` instead.
            See :py:attr:`DatabaseCollection.fields`.
        """
        try:
            return self[collection].fields.values()
        except ValueError:
            return ()

    def set_values(self, collection, document_id, values):
        """
        .. deprecated:: 3.0
            Use ``db_session[collection].update_document(...)`` instead.
            See :py:meth:`DatabaseCollection.update_document`.
        """
        self[collection].update_document(document_id, values)

    def has_document(self, collection, document_id):
        """
        .. deprecated:: 3.0
            Use ``db_session[collection].has_document(...)`` instead.
            See :py:meth:`DatabaseCollection.has_document`.
        """
        return self[collection].has_document(document_id)

    def get_document(self, collection, document_id, fields=None, as_list=False):
        """
        .. deprecated:: 3.0
            Use ``db_session[collection].document(...)`` instead.
            See :py:meth:`DatabaseCollection.document`.
        """
        try:
            collection = self[collection]
        except ValueError:
            return None
        return collection.document(document_id, fields, as_list)

    def get_documents_ids(self, collection):
        """
        .. deprecated:: 3.0
            Use ``db_session[collection].documents_ids(...)`` instead.
            See :py:meth:`DatabaseCollection.documents_ids`.
        """
        try:
            c = self[collection]
        except ValueError:
            return
        yield from c.documents_ids()

    def get_documents(self, collection, fields=None, as_list=False, document_ids=None):
        """
        .. deprecated:: 3.0
            Use ``db_session[collection].documents_ids(...)`` instead.
            See :py:meth:`DatabaseCollection.documents_ids`.
        """
        try:
            c = self[collection]
        except ValueError:
            return
        if document_ids is None:
            yield from c.documents(fields=fields, as_list=as_list)
        else:
            for document_id in document_ids:
                document = c.get(document_id)
                if document is not None:
                    yield document

    def remove_document(self, collection, document_id):
        """
        .. deprecated:: 3.0
            Use ``del db_session[collection][document_id]`` instead.
            See :py:meth:`DatabaseCollection.__delitem__`.
        """
        del self[collection][document_id]

    def add_document(self, collection, document):
        """
        .. deprecated:: 3.0
            Use ``del db_session[collection].add(document)`` instead.
            See :py:meth:`DatabaseCollection.add`.
        """
        self[collection].add(document)

    def filter_documents(self, collection, filter_query, fields=None, as_list=False):
        """
        .. deprecated:: 3.0
            Use ``del db_session[collection].filter(...)`` instead.
            See :py:meth:`DatabaseCollection.filter`.
        """
        yield from self[collection].filter(filter_query, fields=fields, as_list=as_list)


class DatabaseCollection:
    def __init__(self, session, name):
        self.session = session
        self.name = name
        self.catchall_column = self.settings().get("catchall_column", "_catchall")
        self.primary_key = {}
        self.bad_json_fields = set()
        self.fields = {}

    def settings(self):
        return self.session.settings("collection", self.name, {})

    def set_settings(self, settings):
        self.session.set_settings("collection", self.name, settings)

    def document_id(self, document_id):
        if not isinstance(document_id, tuple | list):
            document_id = (document_id,)
        if len(document_id) != len(self.primary_key):
            raise KeyError(
                f"key for table {self.name} requires {len(self.primary_key)} value(s), {len(document_id)} given"
            )
        return document_id

    def update_settings(self, **kwargs):
        settings = self.settings()
        settings.update(kwargs)
        self.set_settings(settings)

    def add_field(
        self, name, field_type, description=None, index=False, bad_json=False
    ):
        """
        Adds a field to the database

        :param collection: Field collection (str, must be existing)

        :param name: Field name (str, must not be existing)

        :param field_type: Field type, in ('string', 'int', 'float', 'boolean', 'date', 'datetime',
                     'time', 'json', 'list_string', 'list_int', 'list_float', 'list_boolean', 'list_date',
                     'list_datetime', 'list_time', 'list_json')

        :param description: Field description (str or None) => None by default

        :param index: Bool to know if indexing must be done => False by default

        :raise ValueError: - If the collection does not exist
                           - If the field already exists
                           - If the field name is invalid
                           - If the field type is invalid
                           - If the field description is invalid
        """
        raise NotImplementedError()

    def remove_field(self, name):
        """
        Removes a field in the collection

        :param collection: Field collection (str, must be existing)

        :param field: Field name (str, must be existing))

        :raise ValueError: - If the collection does not exist
                           - If the field does not exist
        """
        raise NotImplementedError()

    def update_document(self, document_id, partial_document):
        raise NotImplementedError()

    def has_document(self, document_id):
        raise NotImplementedError()

    def document(self, document_id, fields=None, as_list=False):
        raise NotImplementedError()

    def documents(self, fields=None, as_list=False):
        raise NotImplementedError()

    def documents_ids(self):
        yield from (
            i for i in self.documents(fields=tuple(self.primary_key), as_list=True)
        )

    def __iter__(self):
        return self.documents()

    def add(self, document, replace=False):
        raise NotImplementedError()

    def __setitem__(self, document_id, document):
        raise NotImplementedError()

    def _encode_column_value(self, field, value):
        encoding = self.fields.get(field, {}).get("encoding")
        if encoding:
            encode, decode = encoding
            try:
                column_value = encode(value)
            except TypeError:
                # Error with JSON encoding
                column_value = ...
            if column_value is ...:
                column_value = encode(json_encode(value))
                self.bad_json_fields.add(field)
                settings = self.settings()
                settings.setdefault("fields", {}).setdefault(field, {})["bad_json"] = (
                    True
                )
                self.set_settings(settings)
            return column_value
        return value

    def __getitem__(self, document_id):
        return self.document(document_id)

    def __delitem__(self, document_id):
        raise NotImplementedError()

    def parse_filter(self, filter):
        raise NotImplementedError()

    def filter(self, filter, fields=None, as_list=False):
        """
        Iterates over the collection documents selected by filter_query

        Each item yield is a row of the collection table returned

        filter_query can be the result of self.filter_query() or a string containing a filter
        (in this case self.fliter_query() is called to get the actual query)

        :param filter_query: Filter query (str)

                    - A filter row must be written this way: {<field>} <operator> "<value>"
                    - The operator must be in ('==', '!=', '<=', '>=', '<', '>', 'IN', 'ILIKE', 'LIKE')
                    - The filter rows can be linked with ' AND ' or ' OR '
                    - Example: "((({BandWidth} == "50000")) AND (({FileName} LIKE "%G1%")))"

        :param fields: List of fields to retrieve in the document

        :param as_list: If True, document values are returned in a list using
                        fields order

        """
        raise NotImplementedError()

    def delete(self, filter):
        """
        Delete documents corresponding to the given filter

        :param filter_query: Filter query (str)
        """
        raise NotImplementedError()


def json_dumps(value):
    return json.dumps(value, separators=(",", ":"))


_json_encodings = {
    datetime: lambda d: f"{d.isoformat()}ℹdatetimeℹ",
    date: lambda d: f"{d.isoformat()}ℹdateℹ",
    time: lambda d: f"{d.isoformat()}ℹtimeℹ",
    list: lambda l: [json_encode(i) for i in l],  # noqa: E741
    dict: lambda d: {k: json_encode(v) for k, v in d.items()},
}

_json_decodings = {
    "datetime": lambda s: dateutil.parser.parse(s),
    "date": lambda s: dateutil.parser.parse(s).date(),
    "time": lambda s: dateutil.parser.parse(s).time(),
}


def json_encode(value):
    global _json_encodings

    type_ = type(value)
    encode = _json_encodings.get(type_)
    if encode is not None:
        return encode(value)
    return value


def json_decode(value):
    global _json_decodings

    if isinstance(value, list):
        return [json_decode(i) for i in value]
    elif isinstance(value, dict):
        return {k: json_decode(v) for k, v in value.items()}
    elif isinstance(value, str):
        if value.endswith("ℹ"):
            split_value = value[:-1].rsplit("ℹ", 1)
            if len(split_value) == 2:
                encoded_value, decoding_name = split_value
                decode = _json_decodings.get(decoding_name)
                if decode is None:
                    raise ValueError(f'Invalid JSON encoding type for value "{value}"')
                return decode(encoded_value)
    return value


# Obsolete constants kept for backward compatibility with API v2

FIELD_TYPE_STRING = str
FIELD_TYPE_INTEGER = int
FIELD_TYPE_FLOAT = float
FIELD_TYPE_BOOLEAN = bool
FIELD_TYPE_DATE = date
FIELD_TYPE_DATETIME = datetime
FIELD_TYPE_TIME = time
FIELD_TYPE_JSON = dict
FIELD_TYPE_LIST_STRING = list[str]
FIELD_TYPE_LIST_INTEGER = list[int]
FIELD_TYPE_LIST_FLOAT = list[float]
FIELD_TYPE_LIST_BOOLEAN = list[bool]
FIELD_TYPE_LIST_DATE = list[date]
FIELD_TYPE_LIST_DATETIME = list[datetime]
FIELD_TYPE_LIST_TIME = list[time]
FIELD_TYPE_LIST_JSON = list[dict]

ALL_TYPES = {
    FIELD_TYPE_LIST_STRING,
    FIELD_TYPE_LIST_INTEGER,
    FIELD_TYPE_LIST_FLOAT,
    FIELD_TYPE_LIST_BOOLEAN,
    FIELD_TYPE_LIST_DATE,
    FIELD_TYPE_LIST_DATETIME,
    FIELD_TYPE_LIST_TIME,
    FIELD_TYPE_LIST_JSON,
    FIELD_TYPE_STRING,
    FIELD_TYPE_INTEGER,
    FIELD_TYPE_FLOAT,
    FIELD_TYPE_BOOLEAN,
    FIELD_TYPE_DATE,
    FIELD_TYPE_DATETIME,
    FIELD_TYPE_TIME,
    FIELD_TYPE_JSON,
}



from datetime import datetime

schemas = [
    {
        "version": "1.0.0",
        "schema": {
            # A global value
            "last_update": datetime,
            # A single document (i.e. not in a collection)
            "dataset": {
                "directory": str,
                "schema": "str",
            },
            # A collection of metadata associated to a path.
            "metadata": [
                {
                    "path": [str, {"primary_key": True}],
                    "subject": str,
                    "time_point": str,
                    "history": list[str],  # contains a list of execution_id
                }
            ],
        },
    }
]
